@@ -1,5 +1,5 @@
 /*****************************************/
-/* Stack Shield v0.5                     */
+/* Stack Shield v0.6                     */
 /* by Vendicator 1999                    */
 /*****************************************/
 /* File: stackshield.c                   */
@@ -13,6 +13,7 @@
 #include <unistd.h>
 
 #include "globalretstack.h"
+#include "addrcallcheck.h"
 
 void usage(char *progname);
 void parsefile(FILE *srcptr, FILE *destptr);
@@ -23,14 +24,17 @@ void doheader(FILE *fileptr);
 void dodetectattackepilog(FILE *fileptr);
 void doanticrashprolog(FILE *fileptr);
 void doanticrashepilog(FILE *fileptr);
+void docall(FILE *fileptr, char *reg);
 
 int buffelements = 256;
 int anticrash = 0;
 int detectattack = 0;
+int addrcallcheck = -1;
 char entrypoint[257];
 
 int prologcount = 0;
 int epilogcount = 0;
+int callcount = 0;
 int buffsize;
 
 void usage(char *progname) {
@@ -41,6 +45,7 @@ void usage(char *progname) {
   fprintf(stderr, "Options:\n");
   fprintf(stderr, "-c prevents runtime errors if more than too much nested calls are executed\n");
   fprintf(stderr, "-d terminate immidiately when a buffer overflow is detected\n");
+  fprintf(stderr, "-f disable check for function pointer based overflows (use in case of errors)\n");
   fprintf(stderr, "-l elements set buffer size to 'elements' number (elements are 4 bytes) (default 256)\n");
   fprintf(stderr, "-e entrypoint specify the program entry point symbol (default main)\n");
 
@@ -120,18 +125,29 @@ void domainprolog(FILE *fileptr) {
 
 void doheader(FILE *fileptr) {
   char label[257];
+  char *header;
   char *block;
   char *buffer;
   char *tmp;
 
-  block=(char *)calloc(strlen(grsheader)+256, 1);
-  buffer=(char *)calloc(strlen(grsheader)+256, 1);
-  strncpy(block, grsheader, strlen(grsheader));
+  if (addrcallcheck) {
+    header=(char *)calloc(strlen(accheader)+strlen(grsheader), 1);
+    strcpy(header, accheader);
+    strcat(header, grsheader);
+  }
+  else {
+    header=(char *)calloc(strlen(grsheader), 1);
+    strcpy(header, grsheader);
+  }
+
+  block=(char *)calloc(strlen(header)+256, 1);
+  buffer=(char *)calloc(strlen(header)+256, 1);
+  strncpy(block, header, strlen(header));
 
   label[256]='\0';
 
   while (tmp=strstr(block, "<BUFFSIZE>")) {
-    strncpy(buffer, tmp, strlen(grsheader)+256);
+    strncpy(buffer, tmp, strlen(header)+256);
     snprintf(label, 256, "%d", buffsize);
     strncpy(tmp, label, strlen(label));
     strncpy(tmp+strlen(label), buffer+strlen("<BUFFSIZE>"),
@@ -212,6 +228,55 @@ void doanticrashepilog(FILE *fileptr) {
   fprintf(fileptr, "%s", block);
 }
 
+void docall(FILE *fileptr, char *reg) {
+  char preg[257];
+  char label[257];
+  char *call;
+  char *block;
+  char *buffer;
+  char *tmp;
+  int i;
+
+  strncpy(preg, reg, strlen(reg)-1);
+  preg[strlen(reg)-1]='\0';
+
+  call=(char *)calloc(strlen(acccall)+256, 1);
+  strcpy(call, acccall);
+
+  block=(char *)calloc(strlen(call)+256, 1);
+  buffer=(char *)calloc(strlen(call)+256, 1);
+  strncpy(block, call, strlen(call));
+
+  callcount++;
+  label[256]='\0';
+
+  while (tmp=strstr(block, "<CALLCOUNT>")) {
+    strncpy(buffer, tmp, strlen(call)+256);
+    snprintf(label, 256, "%d", callcount);
+    strncpy(tmp, label, strlen(label));
+    strncpy(tmp+strlen(label), buffer+strlen("<CALLCOUNT>"),
+      strlen(buffer+strlen("<CALLCOUNT>")));
+    tmp[strlen(label)+strlen(buffer+strlen("<CALLCOUNT>"))]='\0';
+  }
+  call[strlen(call)+255]='\0';
+  strncpy(call, block, strlen(call)+255);
+  
+  for (i=0; i < strlen(call)+256; i++)
+    buffer[i]=block[i]='\0';
+  strncpy(block, call, strlen(call));
+
+  while (tmp=strstr(block, "<REGISTER>")) {
+    strncpy(buffer, tmp, strlen(call)+256);
+    snprintf(label, 256, "%s", preg);
+    strncpy(tmp, label, strlen(label));
+    strncpy(tmp+strlen(label), buffer+strlen("<REGISTER>"),
+      strlen(buffer+strlen("<REGISTER>")));
+    tmp[strlen(label)+strlen(buffer+strlen("<REGISTER>"))]='\0';
+  }
+
+  fprintf(fileptr, "%s", block);
+}
+
 void parsefile(FILE *srcptr, FILE *destptr) {
   char line[257];
 
@@ -249,6 +314,10 @@ void parsefile(FILE *srcptr, FILE *destptr) {
       fprintf(destptr, "%s", line);
       domainprolog(destptr);
     }
+    else if (strstr(line, "call *") && addrcallcheck) {
+      docall(destptr, strstr(line, "call *")+strlen("call *"));
+      fprintf(destptr, "%s", line);
+    }
     else
     fprintf(destptr, "%s", line);
   }
@@ -262,7 +331,7 @@ int main(int argc, char **argv) {
 
   strncpy(entrypoint, "main", 256);
   entrypoint[256]='\0';
-  while ((opt=getopt(argc, argv, "cdl:e:")) != EOF) {
+  while ((opt=getopt(argc, argv, "cdfl:e:")) != EOF) {
     switch (opt) {
       case 'c':
         anticrash=-1;
@@ -270,6 +339,9 @@ int main(int argc, char **argv) {
       case 'd':
         detectattack=-1;
 	break;
+      case 'f':
+        addrcallcheck=0;
+	break;
       case 'l':
         buffelements=atoi(optarg);
 	break;
